from typing import List, Type
from torch.utils.data import Dataset, DataLoader
from torch.nn.parameter import Parameter
from collections.abc import Iterator
from modules.models.abs_module import AbsModule
from PIL.Image import Image as ImageDT
from matplotlib.axes import Axes
from torch import Tensor



from dependencies.core import glob, os, pathlib, torchvision, tempfile
from dependencies.fancy import colored
from dependencies.datatypes import *

class TorchvisionDownloader:
  """
    Downloads a dataset from torchvision.

    Attributes
    ----------
      dataset_name : str
        The name of the dataset to download.
      dir : str
        The directory to download the training data to.
      split_percentage : tuple[float, float, float]
        The percentages of how to split the data set into three parts.

    Methods
    -------
      call : None
        Downloads the dataset.

  """
  def __init__(self, dir: str, dataset: str,
               split_percentage: tuple[float, float, float] = (0.65, 0.25, 0.1),
               **dataset_kwargs
               ):
    """
      Constructor

      Parameters
      ----------
        test_dir : str, optional
          Path to the directory to save the split of `./train` and `./test`. The default is './data'.
        dataset : str, optional
          Name of the dataset. The default is 'MNIST'.
        split_percentage : int, optional
          Percentage of the dataset to be used for training. The default is 70.

      Returns
      -------
        None
    """
    self.__dataset : Dataset[list[ImageDT]]
    self.__tmp_path = tempfile.gettempdir()
    self.dataset_name = dataset
    self.dataset_kwargs = dataset_kwargs
    self.dataset_kwargs["root"] = self.__tmp_path
    self.dataset_kwargs["download"] = True
    self.split_percentage = split_percentage
    self.__labels = []

    self.__root_dir = dir
    self.__train_dir = pathlib.Path(dir, 'train')
    self.__valid_dir = pathlib.Path(dir, 'valid')
    self.__test_dir = pathlib.Path(dir, 'test')

    if not pathlib.Path(self.__root_dir).exists():
      os.mkdir(self.__root_dir)
      os.mkdir(self.__train_dir)
      os.mkdir(self.__valid_dir)
      os.mkdir(self.__test_dir)
    else:
      if self.__train_dir.exists() or self.__valid_dir.exists() or self.__test_dir.exists():
        self.__clear_tmp()
        raise Exception(f"{self.__root_dir} directory is not empty.")
      else:
        os.mkdir(self.__train_dir)
        os.mkdir(self.__valid_dir)
        os.mkdir(self.__test_dir)

  def call(self, forced: bool = False) -> None:
    """
      Downloads the dataset from torch-vision.

      Parameters
      ----------
        forced : bool, optional
          Force the download of the dataset. The default is False.

      Returns
      -------
        None
    """
    print(colored(' Downloading ... 🌑 ', 'blue', attrs=['bold']), end='\r')

    try:
      self.__dataset = getattr(torchvision.datasets, self.dataset_name)(**self.dataset_kwargs)
      self.__dataset.idx_to_class = {val:key for key, val in self.__dataset.class_to_idx.items()}
    except Exception:
      if not forced:
        raise Exception("Target dataset cannot be downloaded!, please try another one.")

    print('\r', end='\r')
    print(colored(' Processing ...  🌕 ', 'yellow', attrs=['bold']), end='\r')
    self.__iterate_dataset(self.__write_images)
    self.__write_labels()
    self.__clear_tmp()
    print('\r', end='\r')
    print(colored(' Ready           💚', 'green', attrs=['bold']))

  def __clear_tmp(self):
    """
      Clear the tmp directory

      Parameters
      ----------
        None

      Returns
      -------
        None
    """
    filelist = glob.glob(f"{self.__tmp_path}/{self.dataset_name}/raw/*")
    for file in filelist:
      os.remove(file)
    os.rmdir(f"{self.__tmp_path}/{self.dataset_name}/raw")
    os.rmdir(f"{self.__tmp_path}/{self.dataset_name}")

  def __idx_to_img(self, idx):
    """
      Convert the index to the image filename

      Parameters
      ----------
        idx : int
            The index of the image

      Returns
      -------
        str
          The filename for the image
    """
    return f"{idx}.jpg"

  def __iterate_dataset(self, cb_fn):
    """
      Iterates over the dataset and calls the callback function

      Parameters
      ----------
        cb_fn : function
          Callback function to be called on each image

      Returns
      -------
        None
    """
    idx = 0
    total = len(self.__dataset) # pyright: reportGeneralTypeIssues=false
    max_train_idx = int(total * self.split_percentage[0])
    max_valid_idx = int(total * (self.split_percentage[0] + self.split_percentage[1]))

    for img, label in self.__dataset:
      label = self.__custom_label_mapping(label)

      if idx < max_train_idx:
        cb_fn(img, idx, self.__train_dir)
      elif idx < max_valid_idx:
        cb_fn(img, idx, self.__valid_dir)
      else:
        cb_fn(img, idx, self.__test_dir)

      self.__labels.append(f"{self.__idx_to_img(idx)},{label}")
      idx+=1

  def __write_images(self, img, idx, path):
    """
      Write the image to the specified path transforming the .mat file into images.

      Parameters
      ----------
        img : torch.Tensor
          The image tensor
        idx : int
          The index of the dataset, that will be transformed into the filename of the image
        path : str
          The path where the image will be written

      Returns
      -------
        None
    """
    file = pathlib.Path(path, self.__idx_to_img(idx))
    img.save(file.as_posix())

  def __custom_label_mapping(self, label):
    """
      Custom label mapping for the dataset words to an index.

      Parameters
      ----------
        label : str
          The label to be mapped.

      Returns
      -------
        int
          The integer that represents the mapping.
    """
    if hasattr(self.__dataset, 'idx_to_class'):
      return self.__dataset.idx_to_class[label]
    return None

  def __write_labels(self):
    """
      Write the labels to a CSV files from the test and train directories.

      Parameters
      -------
        None

      Returns
      -------
        None
    """
    path = pathlib.Path(self.__root_dir, 'labels.csv')
    with path.open(mode = 'w', encoding='utf-8') as file:
      for label in self.__labels:
        file.write(label)
        file.write('\n')
    file.close()



from dependencies.fancy import colored, spinner
from dependencies.core import multiprocessing, time, abstractmethod, ABCMeta
from libs import colorize_diff

class PrettifyComparer(metaclass=ABCMeta):
  """
    Prettify the comparer by adding spinner, and nice diff output.
  """
  BAR_LENGTH = 55
  LOADING_ICONS = ['⠋','⠙','⠹','⠸','⠼','⠴','⠦','⠧','⠇','⠏']
  WIDTH = 50

  @abstractmethod
  def __init__(self):
    """
      Implementing this __init__ **only** for typing purposes. The child class, shall
      not call this method directly.
    """
    self.models: list
    self.names: list[str]

  def __str__(self) -> str:
    """
      Return a string representation of the comparer.
    """
    last = len(self.models) - 2
    result = ""
    for i, _ in enumerate(self.models[:-1]):
      result += colorize_diff(self.models[i], self.models[i+1],
                              headers=(self.names[i], self.names[i+1]), omit_changes=True
                              )
      if i < last:
        result += '\n '+ colored('-' * self.WIDTH, 'black') +'\n\n'

    return  result

  def _terminate_loading(self) -> None:
    """
      Terminate the loading spinner. Made a method to be called in the child class.
    """
    print(f"\r{' ':>55}\r", end='')
    self.__loading.terminate();

  def _print_results(self, results: list[str]) -> None:
    """
      Print the results and compare them. This will also terminate the spinner.
    """
    self._terminate_loading()
    # print(results)
    last = len(self.models) - 2
    for i, _ in enumerate(self.models[:-1]):
      result = colorize_diff(results[i], results[i+1],
                             headers=(self.names[i], self.names[i+1]), omit_changes=True
                             )
      if i < last:
        result += '\n '+ colored('-' * self.WIDTH, 'black') +'\n\n'
      print(result)

  def __infinite_spinner(self) -> None:
    """
      Print a loading spinner.
    """
    s = spinner(icons=[' ' + colored(i, 'yellow')* self.WIDTH for i in self.LOADING_ICONS])
    while True:
      next(s)
      time.sleep(0.1)

  def _loading(self) -> None:
    """
      Start parallel loading spinner.
    """
    self.__loading = multiprocessing.Process(target=self.__infinite_spinner)
    self.__loading.start()



from dependencies.core import click, torch
from dependencies.fancy import colored
from .shared import labels_reader
from modules import models
from features import Trainer
from modules.transforms.image_transforms import ImageTransforms

@click.command()
@click.pass_context
@click.argument("model_file", type=click.Path(file_okay=True, exists=True, readable=True), required=1)
@click.argument("data_dir", type=click.Path(exists=True, dir_okay=True, readable=True), required=1)
@click.argument("categories-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.option("-b", "--batch-size", type=int, default=16, help="Image loader batch size")
@click.option("-e", "--epochs", default=3, type=click.INT)
@click.option("-l", "--learning-rate", default=round(1/137, 6), type=click.FLOAT)
@click.option("-m", "--momentum", type=click.FLOAT)
@click.option("-p", "--pretend", default=False, type=click.BOOL, is_flag=True)
def train(_, model_file, data_dir, categories_file, batch_size, epochs, learning_rate, momentum, pretend):
  """
    Train a given architecture with a data directory containing a '/validate' and '/train' subfolder
    each with the images files and a `labels.csv` file.
  """
  labels: dict = labels_reader(categories_file, _sorted=False) # pyright: ignore [reportGeneralTypeIssues]

  data = torch.load(model_file)
  model = getattr(models, data['name'])(**{"data": data})

  if pretend:
    print(colored(' Results of the training will not saved, since we are just pretending\n', 'yellow'))

  if momentum is not None:
    trainer = Trainer(model, ImageTransforms, learning_rate = learning_rate, momentum = momentum)
  else:
    trainer = Trainer(model, ImageTransforms, learning_rate = learning_rate)

  trainer.train(data_dir, labels, batch_size, epochs)

  if not pretend:
    model.save(model_file)



"""
  A CLI to download, train, test, predict and compare an image classifiers.

  Supporting mostly all torch-vision neural networks and datasets.

  This will also identify cute 🐱 or a fierce 🐶, also flowers
  or what type of 🏘️ you should be.
"""

from dependencies.core import click
from commands import download, create, train, test, info, predict, compare

__author__ = "aemonge"
__copyright__ = "aemonge"
__license__ = "MIT"
__name__ = "alicia"
__version__ = "0.0.9"

@click.group()
@click.pass_context
@click.option("-v", "--verbose", default=False, is_flag=True, type=click.BOOL)
@click.option("-g", "--gpu", default=False, is_flag=True, type=click.BOOL)
@click.version_option(version=__version__, package_name=__name__)
def call(ctx, verbose, gpu):
  """
    A CLI to download, train, test, predict and compare an image classifiers.

    Supporting mostly all torch-vision neural networks and datasets.

    This will also identify cute 🐱 or a fierce 🐶, also flowers
    or what type of 🏘️ you should be.
  """
  ctx.ensure_object(dict)
  ctx.obj['verbose'] = verbose
  ctx.obj['use_gpu'] = gpu

call.add_command(download)
call.add_command(create)
call.add_command(train)
call.add_command(test)
call.add_command(predict)
call.add_command(info)
call.add_command(compare)

if __name__ == '__main__':
  # ^  This is a guard statement that will prevent the following code from
  #    being executed in the case someone imports this file instead of
  #    executing it as a script.
  #    SEE: https://docs.python.org/3/library/__main__.html
  pass



from dependencies.core import click, inspect
from .shared import labels_reader
from modules import models
MODELS_NAMES = [ name for name, _ in inspect.getmembers(models, predicate=inspect.isclass) ]

@click.command()
@click.pass_context
@click.argument("architecture", default=MODELS_NAMES[0], type=click.Choice(MODELS_NAMES), required=1)
@click.argument("categories-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.argument("save-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.option('-d', '--dropout', type=click.FLOAT, default=0.0)
@click.option('-i', '--input-size', type=click.INT, default=28)
def create(_, architecture: str, categories_file: str, save_file: str, input_size: int, dropout: float) -> None:
  """
    Creates a new model for a given architecture.
    This will generate a .pth file to use later to train, test, and evaluate the model.
  """
  sorted_labels = labels_reader(categories_file)
  model = getattr(models, architecture)(**{
    "labels": sorted_labels, "input_size": input_size, "dropout": dropout
  })

  model.save(save_file)
  print('💚')



from dependencies.core import click
from commands.compare_info import diff_info
from commands.compare_step_speed import step_speed
from commands.compare_accuracy import accuracy

@click.group()
@click.pass_context
def compare(_):
  """
    Compare the info, accuracy, and step speed two (or more by pairs) trained models on a test set.
  """
  pass

compare.add_command(diff_info)
compare.add_command(accuracy)
compare.add_command(step_speed)



# Fancy,
from dependencies.core import time, glob, randrange, Image, os, np
from dependencies.fancy import spinner, loading_bar, colored, plt, plotext
from PIL.Image import Image as ImageDT
from matplotlib.axes import Axes

class PrettyTrain:
  """
    A parent class that prints the trainer data and analytics
      nicely and colored.

    Attributes
    ----------
      spinners_ : dict
        A dictionary of spinners, with different colors.

    Methods
    -------
      _loading(ix, total, step: str = 'train')
        Prints a loading bar.
      _spin(step: str)
        Prints a spinner.
      _print_train_header(self, epochs, batch_size, train_loader_count, validate_loader_count)
        Prints the training header.

  """
  BAR_LENGTH = 55
  LOADING_ICONS = ['⠋','⠙','⠹','⠸','⠼','⠴','⠦','⠧','⠇','⠏']

  def __init__(self):
    """
      Constructor. This is meant to be a parent Class
    """
    self.__enable_spinners = False # Loading bar seams enough
    if self.__enable_spinners:
      self._spinners_ = {
        'train': spinner(icons=[colored(i, 'blue') for i in self.LOADING_ICONS]),
        'test':  spinner(icons=[colored(i, 'yellow') for i in self.LOADING_ICONS]),
        'valid': spinner(icons=[colored(i, 'green') for i in self.LOADING_ICONS])
      }

  def __get_step_color__(self, step: str = 'train') -> str:
    """
      Gets the color of a step.

      Parameters
      ----------
        step : str
          The step to get the color of.

      Returns
      -------
        color : str
          The color of the step.
    """
    color = ""
    match step:
      case 'train':
        color = 'blue'
      case 'valid':
        color = 'green'
      case 'test':
        color = 'yellow'
    return color

  def _loading(self, ix: int, total: int, step: str = 'train') -> None:
    """
      Prints the loading bar.

      Parameters
      ----------
        ix: int
          The index of the loading bar.
        total: int
          The total number of items.
        step: str
          The step(color) of the loading bar.

      Returns:
      --------
        None
    """
    color = self.__get_step_color__(step)
    loading_bar(
      ix, total=total, label='  ', bar_length=self.BAR_LENGTH, show_percentage=True, icon=colored('⠿', color)
    )

  def _spin(self, step: str = 'train') -> None:
    """
      Prints a spinner.

      Parameters:
      -----------
        step: str
          The step(color) of the spinner.

      Returns:
      --------
        None
    """
    if self.__enable_spinners:
      next(self._spinners_[step])

  def __backspace__(self, hard: bool = False):
    """
      Prints a backspace character.

      Parameters:
      -----------
        hard: bool
          Whether to print a hard or soft backspace character.

      Returns:
      --------
        None
    """
    if hard:
      print('\033[F\033[K', end='') # back prev line and clear
    else:
      print('\r', end='\r')

  def _print_train_header(self, epochs:int, batch_size:int,
                          train_loader_count:int, validate_loader_count:int) -> None:
    """
      Prints the header of the training.

      Parameters:
      -----------
        epochs:int
          The number of epochs.
        batch_size:int
          The batch size.
        train_loader_count:int
          The number of train loader.
        validate_loader_count:int
          The number of validate loader.

      Returns:
      --------
        None

    """
    momentum_str = ""
    if hasattr(self, 'momentum') and self.momentum is not None:
      momentum_str = f",\tMomentum: {self.momentum}"

    print(f" Epochs: {epochs},\tBatch Size: {batch_size},\tLearning rate: {self.learning_rate}{momentum_str}\n",
          f"Items: [training: \"{train_loader_count:,}\" ,\tvalidation: \"{validate_loader_count:,}\"]\n")

  def _print_test_header(self, batch_size:int, count:int) -> None:
    """
      Prints the header of the testing.

      Parameters:
      -----------
        batch_size:int
          The batch size.
        count:int
          The total number of items.

      Returns:
      --------
        None

    """
    print(f" Batch Size: {batch_size},\tItems : {count}\n")

  def _print_step_header(self, epochs:int, epoch:int) -> None:
    """
      Prints the header of a step.

      Parameters:
      -----------
        epochs:int
          The number of epochs.
        epoch:int
          The epoch number.

      Returns:
      --------
        None
    """
    print(f"   Epoch: {epoch + 1}/{epochs} ({colored('traning', 'blue')} and {colored('validating', 'green')})")

  def _print_step(self, epoch:int, epochs:int, start_time:int, time_count:int,
                  tr_loss:float, vd_loss:float, vd_correct:int,
                  validate_loader_count:int, train_loader_count:int) -> float:
    """
      Prints a step.

      Parameters:
      -----------
        epoch : int
          The epoch number.
        epochs: int
          The number of epochs.
        start_time: int
          The start time.
        time_count: int
          The time count.
        tr_loss: float
          The training loss.
        vd_loss: float
          The validation loss.
        vd_correct: int
          The validation accuracy.
        validate_loader_count:int
          The number of validate loader.
        train_loader_count:int
          The number of train loader.

      Returns:
      --------
        : float
          The time passed from last step, used to track time
    """

    training_loss = tr_loss / train_loader_count
    validate_loss = vd_loss / validate_loader_count

    time_now = time.time()
    time_count += time_now - start_time
    time_f = time.strftime("%H:%M:%S", time.gmtime(time_now - start_time))
    acc = vd_correct * 100 / validate_loader_count
    accuracy_f = colored(f"{acc:.4f}%", ('blue' if acc > 90 else ('green' if acc > 70 else 'red')))

    print(f"   Epoch: {epoch + 1}/{epochs}\n",
          f"     Losses [ training: {training_loss:.6f}, validate: {validate_loss:.6f} ]{' ':>20}\n",
          f"     Accuracy: {accuracy_f}, Time: {time_f}{' ':>30}\n",
    )

    return time_now

  def _print_total(self, correct, count, start_time) -> None:
    """
      Prints the total.


      Parameters:
      -----------
        correct: int
          The number of correct items
        count: int
          The total number of items.
        start_time: int
          The start time.

      Returns:
      --------
        None
    """
    time_f = time.strftime("%H:%M:%S", time.gmtime(time.time() - start_time))
    acc = (correct * 100 / count) #:.2f
    accuracy_f = colored(f"{acc:.3f}%", ('blue' if acc > 90 else ('green' if acc > 70 else 'red')))

    print(f" Accuracy: {accuracy_f}, Time: {time_f}")

  def _print_t_step(self, start_time, t_correct, test_loader_count) -> None:
    """
      Prints a test step.


      Parameters:
      -----------
        start_time: int
          The start time.
        t_correct: int
          The number of correct items.
        test_loader_count: int
          The number of test loader.

      Returns:
      --------
        None
    """
    time_f = time.strftime("%H:%M:%S", time.gmtime(time.time() - start_time))
    acc = (t_correct * 100 / test_loader_count)
    accuracy_f = colored(f"{acc:.3f}%", ('blue' if acc > 90 else ('green' if acc > 70 else 'red')))

    self.__backspace__()
    print(f" Accuracy: {accuracy_f}, Time: {time_f}{' ':>40}")

  def _print_pbs_in_console(self, labels:dict, probs, subplot_idx) -> None:
    """
      Print the probability distribution in console.

      Parameters:
      -----------
        labels: dict
          The labels.
        probs:
          The probability distribution.
        ax: matplotlib.axes.Axes
          The axis to plot.

      Returns:
      --------
        None
    """
    probs = np.round(probs, 2)
    plotext.subplot(subplot_idx + 1, 1)
    plotext.bar(labels, probs, orientation = "h", width = 0.1)

  def _imshow_in_console(self, image:ImageDT, subplot_idx:int, *, title:str, tilted:bool =False)-> None:
    """
      Plot the image.

      Parameters:
      -----------
        image: Image
          The image to plot
        ax: matplotlib.axes.Axes
          The axis to plot.
        title: str
          The title of the Image.
        tilted: bool
          Whether the image title should be vertical.

      Returns:
      --------
        None
    """
    plotext.subplot(subplot_idx + 1, 2)
    plotext.image_plot(image)
    plotext.xlabel(title, xside=1)

  def _print_pbs(self, labels:dict, probs, ax:Axes) -> None:
    """
      Print the probability distribution.

      Parameters:
      -----------
        labels: dict
          The labels.
        probs:
          The probability distribution.
        ax: matplotlib.axes.Axes
          The axis to plot.

      Returns:
      --------
        None
    """
    probs = np.round(probs, 2)

    if ax is None:
      plt.subplot(1, 1)
      plt.barh(labels, probs)
    else:
      ax.barh(labels, probs)

  def _imshow(self, image:ImageDT, *, ax:Axes, title:str, tilted:bool =False)-> Axes:
    """
      Plot the image.

      Parameters:
      -----------
        image: Image
          The image to plot
        ax: matplotlib.axes.Axes
          The axis to plot.
        title: str
          The title of the Image.
        tilted: bool
          Whether the image title should be vertical.

      Returns:
      --------
        None
    """
    if ax is None:
      _, ax = plt.subplots()

    if title is not None:
      if tilted:
        ax.set_ylabel(title.replace(' ', '\n'), fontsize=10)
        ax.yaxis.set_label_position("right")
        ax.tick_params(axis='y')
        ax.set_xticks([])
        ax.set_yticks([])
      else:
        ax.set_title(title)
        ax.axis("off")

    ax.imshow(image, interpolation='nearest')
    return ax

  def show_test_results(self, path: str, labels: dict, count: int, tilted: bool = False) -> None:
    """
      Show the test results.

      Parameters:
      -----------
        path: str
          The path to the test images.
        labels: dict
          The class labels.
        count: int
          The total number of items do show.
        tilted: bool
          Whether the image title should be vertical.

      Returns:
      --------
        None
    """
    all_imgs = [
      file for file in glob.glob(f"{path}/**/*.@(jpg|jpeg|png)", flags=glob.EXTGLOB)
    ]

    _, ax = plt.subplots(count, 2, width_ratios=[4,1])
    if count == 1:
      ax = [ax]

    if tilted:
      plt.subplots_adjust(left=0.22, bottom=0.08, top=0.98, hspace=0.6, wspace=0)
    else:
      plt.subplots_adjust(left=0.22, bottom=0.08, top=0.90, hspace=0.6, wspace=0.35, right=0.86)

    for c_idx in range(count):
      ix = randrange(len(all_imgs))
      img_path = all_imgs[ix]

      image = Image.open(img_path)
      file_name = os.path.basename(img_path)

      probs, classes = self.predict(image, topk=3)
      self._print_pbs(classes, probs, ax=ax[c_idx][0])
      self._imshow(self.transforms['display'](image), title=labels[file_name], ax=ax[c_idx][1], tilted=tilted)

    plt.show()

  def show_test_results_in_console(self, path: str, labels: dict, count: int) -> None:
    """
      Show the test results in console.

      Parameters:
      -----------
        path: str
          The path to the test images.
        labels: dict
          The class labels.
        count: int
          The total number of items do show.
        tilted: bool
          Whether the image title should be vertical.

      Returns:
      --------
        None
    """
    all_imgs = [
      file for file in glob.glob(f"{path}/**/*.@(jpg|jpeg|png)", flags=glob.EXTGLOB)
    ]

    plotext.subplots(count, 2)
    plotext.plot_size(80, 15*count)

    for c_idx in range(count):
      ix = randrange(len(all_imgs))
      img_path = all_imgs[ix]

      image = Image.open(img_path)
      file_name = os.path.basename(img_path)

      probs, classes = self.predict(image, topk=3)
      self._print_pbs_in_console(classes, probs, c_idx)
      self._imshow_in_console(img_path, c_idx, title=labels[file_name])

    plotext.show()
    plotext.clf()



from dependencies.core import torch, abstractmethod, abstract_attribute, ABCMeta, textwrap, sys, time

# TODO: Move this as a library
def sizeof_formated(num, suffix='B'):
    """
      Converts a size in bytes to a human-readable format.

      Parameters:
      -----------
        num: int
          Size in bytes.
        suffix: str
          Suffix to append to the unit.

      Returns:
      --------
        : str
          Human-readable size.
    """
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Yi{suffix}"

class AbsModule(torch.nn.Module, metaclass=ABCMeta):
  @abstract_attribute
  def labels(self) -> list|None:
    pass

  def __call__(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
    """
    return self.forward(x)

  def __str__(self):
    """
      A verbose string representation of the neural network.

      Returns:
      --------
        : str
          labels, features, classifier
    """
    features_str = "\n  ".join(str(self.features).split("\n"))
    labels_str = textwrap.fill(str(self.labels)[1:-1], width=80)
    labels_str = labels_str.replace('\n', '\n' + ' '*4)
    formated_size = sizeof_formated(sys.getsizeof(self))

    if hasattr(self, "classifier"):
      classifier_str = "\n  ".join(str(self.classifier).split("\n"))

    if hasattr(self, 'dropout'):
      meta_str = f"size: {formated_size},\tdropout: {self.dropout},\tinput size:{self.input_size}" + \
        f"\t label count (output): {self.num_classes}"
    else:
      meta_str = f"size: {formated_size},\tinput size:{self.input_size}," + \
        f"\t label count (output): {self.num_classes}"

    training_history_str = ""*4
    for line in self.training_history:
      training_history_str += f"Epochs: {line[0][1]}, "
      training_history_str += f"Batch: {line[1][1]}, "
      training_history_str += f"Items: ({line[2][1]}, {line[2][2]})\n"

      time_f = time.strftime("%H:%M:%S", time.gmtime(line[3][1]))
      training_history_str += ' '*6 + f"Accuracy: {line[4][1]:.2f}%, "
      training_history_str += f"Time: {time_f}\n" + " "*4
    training_history_str = training_history_str[:-(4-1)] # 4 will be a constant

    return f"{self.__repr__()} (\n" + \
      f"  (meta): \n    {meta_str}\n" + \
      f"  (labels): \n    {labels_str}\n" + \
      f"  (features): {features_str}\n" + \
      (f"  (classifier): {classifier_str}\n" if hasattr(self, 'classifier') else '') + \
      (f"  (training history): \n    {training_history_str[:-1]}" if len(self.training_history) > 0 else '') + \
    f")"

  @abstractmethod
  def parameters(self) -> list:
    pass

  def save(self, path: str) -> None:
    """
      Save the neural network.

      Parameters:
      -----------
        path: str
          path to save model

      Returns:
      --------
        None
    """
    obj = {
      'name': self.__repr__()[:-2], # Trim the () ex. Basic()/Basic
      'input_size': self.input_size,
      'labels': self.labels,
      'features': self.features,
      'state_dict': self.state_dict(),
    }

    if hasattr(self, 'training_history'):
      obj['training_history'] = self.training_history
    if hasattr(self, 'dropout'):
      obj['dropout'] = self.dropout
    if hasattr(self, 'classifier'):
      obj['classifier'] = self.classifier

    torch.save(obj, path)

  def load(self, state_dict: dict) -> None:
    """
      Parameters:
      -----------
        path: str
          path to load model

      Returns:
      --------
        None
    """
    self.load_state_dict(state_dict)



from dependencies.core import click, torch
from modules import models

@click.command()
@click.pass_context
@click.argument("model_file", type=click.Path(file_okay=True, exists=True, readable=True), required=1)
def info(_, model_file):
  """
    Display information about a model architecture.
    Hidden layers, the trained time, out put, and features.
  """
  data = torch.load(model_file)
  model = getattr(models, data['name'])(**{"data": data})

  print(model)



from dependencies.core import *
from dependencies.datatypes import *
from libs import PrettyTrain, UnLabeledImageDataset

class Trainer(PrettyTrain):
  """
    Base class for all training

    Attributes
    ----------
      model: AbsModule
        The model to train.
      Transforms: Transforms
        The transformations to apply to the data.
      learning_rate: float
        The learning rate to use.
      momentum: float
        The momentum to use.

    Methods
    -------
      train_step(images: torch.Tensor, labels: torch.Tensor, loss_count: float) -> float
      validation_step(dataloaders: DataLoader, batch_size: int) -> tuple[float, int]
      train(self, data_dir: str, labels: dict, batch_size: int = 64, epochs: int = 1,
            freeze_parameters: bool = False) -> None
      test(self, data_dir: str, labels: dict, batch_size: int = 64, freeze_parameters: bool = False)
      predict(self, image: Pil.Image, topk:int=5)
  """

  def __init__(
    self, model: AbsModule, transforms, learning_rate: float = 1/137,  momentum: float|None = None
  ) -> None:
    """
      Parameters
      ----------
        model: AbsModule
          The model to train.
        transforms: Transforms
          The transformations to apply to the data.
        learning_rate: float
          The learning rate to use.
        momentum: float
          The momentum to use.

      Returns
      -------
        None
    """
    super().__init__()
    self.model = model
    self.learning_rate = learning_rate

    self.criterion = torch.nn.CrossEntropyLoss()
    if momentum is not None:
      self.momentum = momentum
      self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.learning_rate, momentum=self.momentum)
    else:
      self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.learning_rate)
    self.transforms = transforms

  def train_step(self, images: torch.Tensor, labels: torch.Tensor, loss_count: float) -> float:
    """
      Performs a single training step.

      Parameters
      ----------
        images: torch.Tensor
          The images to train on.
        labels: torch.Tensor
          The labels to train with.
        loss_count: float
          The loss count to increment.

      Returns
      -------
        : float
          The loss count.
    """
    self._spin()

    self.optimizer.zero_grad()
    self._spin()

    output = self.model(images)
    self._spin()

    loss = self.criterion(output, labels)
    self._spin()

    loss.backward()
    self._spin()

    self.optimizer.step()
    self._spin()

    loss_count += loss.item()
    self._spin()

    return loss_count

  def validation_step(self, dataloaders: DataLoader, batch_size: int) -> tuple[float, int]:
    """
      Performs a single validation step.

      Parameters
      ----------
        dataloaders: DataLoader
          The dataloaders to use.
        batch_size: int
          The batch size to use.

      Returns
      -------
        : tuple[float, int]
          The loss count and the number of correct predictions.
    """
    ix = 0
    vd_loss = 0.0
    vd_correct = 0
    validate_loader_count = len(dataloaders.dataset) # pyright: reportGeneralTypeIssues=false

    with torch.no_grad():
      self.model.eval()
      self._spin(step='valid')

      for (images, (labels, _)) in iter(dataloaders):
        ix += batch_size * 1
        self._spin(step='valid')

        output = self.model.forward(images)
        self._spin(step='valid')

        loss = self.criterion(output, labels)
        self._spin(step='valid')

        vd_loss += loss.item()
        self._spin(step='valid')

        ps = torch.exp(output)
        _, top_class = ps.topk(1, dim=1)
        equals = top_class == labels.view(*top_class.shape)
        vd_correct += equals.sum().item()
        self._spin(step='valid')

        self._loading(ix, validate_loader_count, step = 'valid')

    self._spin(step='valid')
    self.model.train()
    return vd_loss, vd_correct

  def train(self, data_dir: str, labels: dict, batch_size: int = 64, epochs: int = 1,
            freeze_parameters: bool = False) -> None:
    """
      Trains the model.

      Parameters
      ----------
        data_dir: str
          The path to the data directory.
        labels: dict
          The labels to train with.
        batch_size: int
          The batch size to use.
        epochs: int
          The number of epochs to train for.
        freeze_parameters: bool
          Whether to freeze the parameters.

      Returns
      -------
        None

      Help:
      ----------
        labels['25.jpg'] # -> Str
        category_labels_ids[labels['25.jpg']] # -> Int
    """
    if 'train' not in self.transforms.keys() or 'valid' not in self.transforms.keys():
      raise ValueError('Transforms must be defined and set')

    category_labels_ids = { v:k for k,v in enumerate(self.model.labels)}
    train_ldr = DataLoader(UnLabeledImageDataset(
        f"{data_dir}/train", labels, category_labels_ids, transform = self.transforms['train']
      ), batch_size = batch_size, shuffle=True
    )
    valid_ldr = DataLoader(UnLabeledImageDataset(
        f"{data_dir}/valid", labels, category_labels_ids, transform = self.transforms['valid']
      ), batch_size = batch_size, shuffle=True
    )

    train_loader_count = len(train_ldr.dataset)
    validate_loader_count = len(valid_ldr.dataset)

    self._print_train_header(epochs, batch_size, train_loader_count, validate_loader_count)

    time_count = 0
    vd_correct = 0
    total_time = time.time()
    start_time = total_time

    # Freeze parameters
    if (freeze_parameters):
      for param in self.model.features.parameters():
          param.requires_grad = False

    for epoch in range(epochs):
      tr_loss = 0.0
      ix = 0

      self._print_step_header(epochs, epoch)
      for (images, (labels, _)) in iter(train_ldr):
        ix += batch_size * 1
        tr_loss = self.train_step(images, labels, tr_loss)
        self._loading(ix, train_loader_count)
        if math.isnan(tr_loss):
          raise Exception('Loss has been lost, check parameters')

      else:
        vd_loss, vd_correct = self.validation_step(valid_ldr, batch_size)

        self.__backspace__(hard=True)
        time_now = self._print_step(
          epoch, epochs, start_time, time_count,
          tr_loss, vd_loss, vd_correct,
          validate_loader_count, train_loader_count
        )
        start_time = time_now
    else:
      self._print_total(vd_correct, validate_loader_count, total_time)

    self.model.training_history.append((
      ('epochs', epochs), ('batch size', batch_size), ('items', train_loader_count, validate_loader_count),
      ('time', time.time() - total_time), ('accuracy', (vd_correct * 100 / validate_loader_count))
    ))

  def test(self, data_dir: str, labels: dict, batch_size: int = 64, freeze_parameters: bool = False):
    """
      Tests the model.

      Parameters
      ----------
        data_dir: str
          The path to the data directory.
        labels: dict
          The labels to test with.
        batch_size: int
          The batch size to use.
        freeze_parameters: bool
          Whether to freeze the parameters.

      Returns
      -------
        None
      """
    if 'test' not in self.transforms.keys():
      raise ValueError('Test or valid transforms must be defined and set')

    t_correct = 0
    ix = 0
    start_time = time.time()
    category_labels_ids = { v:k for k,v in enumerate(self.model.labels)}
    test_ldr = DataLoader(UnLabeledImageDataset(
      f"{data_dir}/test", labels, category_labels_ids, transform = self.transforms['valid']
      ), batch_size = batch_size, shuffle=True
    )

    test_loader_count = len(test_ldr.dataset)
    self._print_test_header(batch_size, test_loader_count)

    if freeze_parameters: # TODO: I think isn't needed with torch.no_grad(), investigate
      for param in self.model.parameters():
        param.requires_grad = False

    with torch.no_grad():
      self.model.eval()
      self._spin(step='test')

      for (images, (labels, _)) in iter(test_ldr):
        ix += batch_size * 1
        self._spin(step='test')

        output = self.model(images)
        self._spin(step='test')

        ps = torch.exp(output)
        _, top_class = ps.topk(1, dim=1)
        equals = top_class == labels.view(*top_class.shape)
        t_correct += equals.sum().item()
        self._spin(step='test')

        self._loading( ix, test_loader_count, step = 'test')
      else:
        self._print_t_step(start_time, t_correct, test_loader_count)

  def predict_image(self, image: str, **kwargs) -> tuple[np.array, list[str]]:
    """
      Calls the predict method, by transforming the given image path to a Pil.Image

      Parameters
      ----------
        image: str
          The path to the image to predict.
        topk: int
          The number of predictions to return.

      Returns
      -------
        : tuple[np.array, list[str]]
          The top-k probabilities and the top-k class labels.

    """
    return self.predict(Image.open(image), **kwargs)

  def predict(self, image:ImageDT, topk:int = 5) -> tuple[np.array, list[str]]:
    """
      Predicts the class of an given image.

      Parameters
      ----------
        image: Pil.Image
          The image to predict.
        topk: int
          The number of predictions to return.

      Returns
      -------
        : tuple[np.array, list[str]]
          The top-k probabilities and the top-k class labels.
    """
    with torch.no_grad():
      self.model.eval()
      tensor_img = self.transforms['test'](image)
      logps = self.model(tensor_img.unsqueeze(0)) # from 3d to 4d [ introducing a batch dimension ]
      ps = logps[0] # Return to 3D [ no batches again ]
      ps_val, ps_idx = ps.topk(topk)

      label_predictions = [self.model.labels[x] for x in ps_idx.numpy()]
      probabilities = [ 100 * (1 / (1 + math.exp(-x))) for x in ps_val]

      return probabilities, label_predictions



from .pretty_train import PrettyTrain
from .un_labeled_image_dataset import UnLabeledImageDataset
from .image_transforms import ImageToMatPlotLib
from .shape_transforms import UnShapetransform, Reshapetransform
from .colorized_diff import colorize_diff
from .prettify_comparer import PrettifyComparer



from dependencies.core import click, torch, os
from .shared import labels_reader
from modules import models as Models
from features import Comparer, Trainer

@click.command()
@click.pass_context
@click.option("-d", "--data-dir", type=click.Path(exists=True, dir_okay=True, readable=True), required=1)
@click.option("-c", "--categories-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.option("-b", "--batch-size", type=int, default=16, help="Image loader batch size")
@click.argument("models_files", nargs=-1, type=click.Path(file_okay=True, exists=True, readable=True), required=1)
def accuracy(_, data_dir, categories_file, batch_size, models_files):
  """
    Accuracy of the models, with the total time given.
    Like running `alicia test` twice with a nice diff.
  """
  labels: dict = labels_reader(categories_file, _sorted=False) # pyright: ignore [reportGeneralTypeIssues]
  models = []
  for model_file in models_files:
    data = torch.load(model_file)
    model = getattr(Models, data['name'])(**{"data": data})
    models.append(model)

  c = Comparer(Trainer, models, names=[ os.path.basename(n) for n in models_files])
  c.accuracy(data_dir, labels, batch_size)



from dependencies.core import torchvision
from libs import ImageToMatPlotLib, UnShapetransform, Reshapetransform

ImageTransforms = {
  "valid": torchvision.transforms.Compose([
    # torchvision.transforms.Grayscale(), # Changes the size to [1, 1, 28, 28] [batch, channels, width, height]
    torchvision.transforms.Resize(28),
    torchvision.transforms.CenterCrop(28),
    torchvision.transforms.ToTensor(),
    # torchvision.transforms.Normalize((0.5,), (0.5,)),
    # Reshapetransform((-1, )),
    # torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
  ]),
  "display": torchvision.transforms.Compose([
    torchvision.transforms.Grayscale(),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.5,), (0.5,)),
    UnShapetransform((-1, )),
    ImageToMatPlotLib((-1, )),
  ]),
  "test": torchvision.transforms.Compose([
    # torchvision.transforms.Grayscale(), # Changes the size to [1, 1, 28, 28] [batch, channels, width, height]
    torchvision.transforms.Resize(28),
    torchvision.transforms.CenterCrop(28),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.5,), (0.5,)),
    # Reshapetransform((-1, )),
    # torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
  ]),
  "train": torchvision.transforms.Compose([
    # torchvision.transforms.Grayscale(), # Changes the size to [1, 1, 28, 28] [batch, channels, width, height]
    torchvision.transforms.Resize(28),
    torchvision.transforms.CenterCrop(28),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize((0.5,), (0.5,)),
    # Reshapetransform((-1, )),
    # torchvision.transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
  ])
}



from modules.transforms.image_transforms import ImageTransforms
from dependencies.core import contextlib, io
from dependencies.datatypes import AbsModule, Type
from libs import PrettifyComparer
from features import Trainer

class Comparer(PrettifyComparer):
  """
    Compare several neural networks performances.
    Test their accuracy or their training and validation loss.

    Methods:
    --------

  """

  def __init__(self, Trainer: Type[Trainer], models: list[AbsModule], *, names: list[str]):
    """
      Initialize

      Parameters
      ----------
        Trainer : Type[Trainer]
          Class of the trainer to use.
        models : list[AbsModule]
          List of neural networks to compare.
        names : list[str]
          Names of the models to compare.

      Returns:
      --------
        None
    """
    self.Trainer = Trainer
    self.models = models
    self.names = names

  def accuracy(self, *args, **kwargs) -> None:
    """
      Compute the overall test accuracy of the models and compare them.

      Parameters:
      -----------
        args:
          Arguments to pass to the `test` method of each model.
        kargs:
          Keyword Arguments to pass to the `train` method of each model.

      Returns:
      --------
        None
    """
    self._loading()

    results = []
    for model in self.models:
      t = self.Trainer(model, ImageTransforms)
      with contextlib.redirect_stdout(io.StringIO()) as f:
        t.test(*args, **kwargs)
        results.append(f.getvalue())

    self._print_results(results)

  def training(self, data_dir, labels, batch_size, **kwargs) -> None:
    """
      Start parallel training of the models and compare their training and validation loss, time and accuracy.

      Parameters:
      -----------
        data_dir : str
          Path to the directory containing the images in /valid and /train
        labels : str
          Path to the labels file
        batch_size : int
          Batch size
        kargs:
          Keyword Arguments to pass to the `Trainer()` of each model.

      Returns:
      --------
        None
    """
    self._loading()

    results = []
    try:
      for model in self.models:
        t = self.Trainer(model, ImageTransforms, **kwargs)
        with contextlib.redirect_stdout(io.StringIO()) as f:
          t.train(data_dir, labels, batch_size, 1)
          results.append(f.getvalue())
    except Exception as e:
      self._terminate_loading()
      raise e

    self._print_results(results)



import torch
from torch import Tensor


class UnShapetransform(object):
  """
    This class is used to transform a tensor from the original shape.
  """
  def __init__(self, shape:tuple):
    """
      Parameters:
      -----------
        shape: tuple
          The shape of the tensor.
    """
    self.shape = shape

  def __call__(self, tensor:Tensor):
    """
      Transform a tensor from the original shape.

      Parameters:
      -----------
        tensor: torch.Tensor
          The tensor to be transformed.

      Returns:
      --------
        torch.Tensor
          The transformed tensor.
    """
    return torch.reshape(tensor, (1, 28, 28))

class Reshapetransform(object):
  """
    This class is used to transform a tensor's shape
  """
  def __init__(self, shape:tuple):
    """
      Parameters:
      -----------
        shape: tuple
          The shape of the tensor.
    """
    self.shape = shape

  def __call__(self, tensor:Tensor):
    """
      Transform a tensor's shape.

      Parameters:
      -----------
        tensor: torch.Tensor
          The tensor to be transformed.

      Returns:
      --------
        torch.Tensor
          The transformed tensor.
    """
    return torch.reshape(tensor, self.shape)
    # return image.view(image.shape[0], -1)



from dependencies.core import glob, pathlib, torch, torchvision, Image
from dependencies.datatypes import Dataset

BATCH_SIZE = 4

class UnLabeledImageDataset(Dataset):
  """
    Custom dataset for loading data from a folder, and labels form the folder file `labels.csv`

    Attributes
    ----------
      main_dir : str
        Path to the folder containing the images
      transform: torchvision.transforms.Compose
        Transform to be applied to the images

    Methods
    -------
  """

  def __init__( self, main_dir, labels: dict, labels_ids: dict, *, transform: torchvision.transforms.Compose):
    """
      Parameters
      ----------
        main_dir: str
          Path to the folder containing the images
        labels: dict
          The Clases fetch from the labels files, which the images will be tested to match {'cat': 0, 'dog': 1}
        labels_ids: dict
          The ids of the labels, which the images will be tested to match {'cat': 0, 'dog': 1}
        transform: torchvision.transforms.Compose
          Transform to be applied to the images

      Returns
      -------
        None
    """
    self.__imgs = [file for file in glob.glob(f"{main_dir}**/*.@(jpg|jpeg|png)", flags=glob.EXTGLOB)]
    self.main_dir = main_dir
    self.transform = transform
    self.__labels = labels
    self.__labels_ids = labels_ids

  def __len__(self):
    """
      Returns the number of samples in the dataset.

      Parameters
      ----------
        None

      Returns
      -------
        int
          The number of samples in the dataset.
    """
    return len(self.__imgs)

  def __getitem__(self, idx) -> tuple[Image.Image|torch.Tensor, tuple[torch.Tensor, str]]:
    """
      Returns a sample from the dataset.

      Parameters
      ----------
        idx : integer
          Index of the image to be loaded.

      Returns
      -------
        image,[class_tensor, file_name] : tuple(torch.Tensor, tuple(torch.Tensor, str))
          Batch image, class label (if any, empty tensor when no class) and its file name.
    """
    file = pathlib.Path(self.__imgs[idx])
    image = Image.open(file.as_posix()).convert("RGB")

    image = self.transform(image)
    class_name = self.__labels[file.name]
    class_id = self.__labels_ids[class_name]

    return image, (torch.tensor(class_id), file.name)



from dependencies.core import click, torch, os
from modules import models as Models
from features import Comparer, Trainer

@click.command()
@click.pass_context

@click.argument("models_files", nargs=-1)
def diff_info(_, models_files):
  """
    Compares the models described in the given files.
  """
  models = []
  for model_file in models_files:
    data = torch.load(model_file)
    model = getattr(Models, data['name'])(**{"data": data})
    models.append(model)

  c = Comparer(Trainer, models, names=[ os.path.basename(n) for n in models_files])
  print(c)



DEBUG_MODE = True

from .basic import Basic
from .dummy import Dummy
from .squeezenet import Squeezenet
from .alexnet import Alexnet
from .mnasnet import Mnasnet

if DEBUG_MODE:
  from .elemental import Elemental



from dependencies.core import difflib
from dependencies.fancy import colored

def colorize_diff(a, b, *, headers: tuple[str,str], omit_changes: bool = False) -> str:
  """
    Colorize the differences between two strings.

    Parameters:
    -----------
      a : str
        The first string.
      b : str
        The second string.
      headers : tuple[str, str]
        The headers to use as title for the diff

    Returns:
    --------
      : str
        The colorized diff.
  """
  diffs = difflib.ndiff(str(a).splitlines(keepends=True), str(b).splitlines(keepends=True))

  output = ""
  if headers:
    output = f"  {colored(headers[0], 'blue')}, {colored(headers[1], 'green')}\n\n"

  for line in diffs:
    if line.startswith('+'):
      output += '  ' + colored(line[2:], 'green')
    elif line.startswith('-'):
      output += '  ' + colored(line[2:], 'blue')
    elif line.startswith('?'):
      if not omit_changes:
        output += '  ' + colored(line[2:], 'yellow')
    else:
      output += line

  return output



from torch import Tensor

class ImageToMatPlotLib(object):
  """
    PyTorch tensors assume the color channel is the first dimension
      but matplotlib assumes is the third dimension
  """
  def __init__(self, shape:tuple):
    """
      Parameters:
      -----------
        shape: tuple
          The shape of the tensor.
    """
    self.shape = shape

  def __call__(self, tensor:Tensor):
    """
      Transform a tensor from the original shape to MatPlotLib image.

      Parameters:
      -----------
        tensor: torch.Tensor
          The tensor to be transformed.

      Returns:
      --------
        torch.Tensor
          The transformed tensor.
    """
    return tensor.numpy().transpose((1, 2, 0))



from dependencies.core import click, torch
from .shared import labels_reader
from modules import models
from features import Trainer
from modules.transforms.image_transforms import ImageTransforms

@click.command()
@click.pass_context
@click.argument("model_file", type=click.Path(file_okay=True, exists=True, readable=True), required=1)
@click.argument("data_dir", type=click.Path(exists=True, dir_okay=True, readable=True), required=1)
@click.argument("categories-file", type=click.Path(file_okay=True, writable=True), required=1)
@click.option("-b", "--batch-size", type=int, default=16, help="Image loader batch size")
@click.option('-c', '--console-plot', default=False, type=click.BOOL, is_flag=True)
@click.option('-h', '--h-title', default=False, type=click.BOOL, is_flag=True, help="Show the title horizontally")
@click.option("-n", "--n-images-test", default=0, type=click.INT,
  help="Increment the number of images to display with the class bar chart and images preview"
)
def test(_, model_file, data_dir, categories_file, batch_size, console_plot, h_title, n_images_test):
  """
    Test a pre trained model. It will look for the `./test` folder inside the data directory.
    A file named `./labels.csv` should exist in the root directory of the data folder.
  """
  labels: dict = labels_reader(categories_file, _sorted=False) # pyright: ignore [reportGeneralTypeIssues]

  data = torch.load(model_file)
  model = getattr(models, data['name'])(**{"data": data})

  trainer = Trainer(model, ImageTransforms)
  trainer.test(data_dir, labels, batch_size)
  if n_images_test > 0:
    if console_plot:
      trainer.show_test_results_in_console(data_dir, labels, n_images_test)
    else:
      trainer.show_test_results(data_dir, labels, n_images_test, h_title)



# pyright: reportUnusedImport=false
from commands.create import create
from commands.download import download
from commands.info import info
from commands.predict import predict
from commands.test import test
from commands.train import train
from commands.compare import compare



from dependencies.core import torch, torchvision
from dependencies.datatypes import Parameter, Iterator
from .abs_module import AbsModule

class Mnasnet(AbsModule, torchvision.models.MNASNet):
  """
    A basic neural network module. With randomly selected features.

    Attributes:
    -----------
      labels: list
        A list of labels.
      features: list
        A list of features.
      classifier: torch.nn.Module
        A classifier.
      input_size: int
        The input size.
      num_classes: int
        The number of classes.
      dropout: float
        The dropout probability.

    Methods:
    --------
      forward(self, x: torch.Tensor) -> torch.Tensor
        A forward pass of the neural network.
      load(self, path: str) -> None
        A load of the neural network.
      save(self, path: str) -> str
        A save of the neural network.
      create(self, input_size: int, num_classes: int, dropout: float) -> None
        Re creates the neural network.
  """

  def __call__(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
    """
    return self.forward(x)

  def __repr__(self):
    """
      A string representation of the neural network.

      Returns:
      --------
        : str
          A string representation 'Basic()'.
    """
    return 'Squeezenet()'

  def parameters(self) -> Iterator[Parameter]:
    """
      Get the parameters of the neural network.

      Returns:
      --------
        Iterator[Parameter]
    """
    return self.features.parameters()

  def __init__(self, *, data: dict|None = None, labels = [],
               input_size: int = 28, dropout: float = 0.0) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
        data: dict
          A dictionary containing the data, to load the network though the pth file.
        labels: list
          A list of labels.
        input_size: int
          The input size.
        dropout: float
          The dropout probability.
    """
    if data is None:
      super().__init__(alpha=1.0, num_classes=len(labels), dropout=dropout)
      self.input_size = input_size
      self.labels = labels
      self.training_history = []
      self.num_classes = len(labels)
      self.features = self.layers
    else:
      if 'dropout' in data:
        self.dropout = data['dropout']
      self.labels = data['labels']
      super().__init__(alpha=1.0, num_classes=len(self.labels), dropout=dropout)

      self.input_size = data['input_size']
      self.features = self.layers
      self.features = data['features']
      if 'training_history' in data:
        self.training_history = data['training_history']
      if 'dropout' in data:
        self.dropout = data['dropout']
      if 'classifier' in data:
        self.classifier = data['classifier']



from dependencies.core import click, torch
from modules import models
from modules.transforms.image_transforms import ImageTransforms
from features import Trainer

@click.command()
@click.pass_context
@click.argument("model_file", type=click.Path(file_okay=True, exists=True, readable=True), required=1)
@click.argument("image", type=click.Path(exists=True, file_okay=True, readable=True), required=1)
@click.option("-k", "--top-k", default=1, type=click.INT, help="Show the top-k most likely categories.")
def predict(_, model_file, image, top_k):
  """
    Predict images using a pre trained model, for a given folder and with a categories file.
  """
  data = torch.load(model_file)
  model = getattr(models, data['name'])(**{"data": data})

  trainer = Trainer(model, ImageTransforms)
  probs, labels = trainer.predict_image(image, topk= top_k)

  if top_k == 1:
    print(f"  \"{labels[0]}\"")
  else:
    for i in range(top_k):
      print(f"  {labels[i]}:\t{probs[i]:04.1f}%")



from dependencies.core import torch
from dependencies.datatypes import Parameter, Iterator
from .abs_module import AbsModule

class Basic(AbsModule):
  """
    A basic neural network module. With randomly selected features.

    Attributes:
    -----------
      labels: list
        A list of labels.
      features: list
        A list of features.
      classifier: torch.nn.Module
        A classifier.
      input_size: int
        The input size.
      num_classes: int
        The number of classes.
      dropout: float
        The dropout probability.

    Methods:
    --------
      forward(self, x: torch.Tensor) -> torch.Tensor
        A forward pass of the neural network.
      load(self, path: str) -> None
        A load of the neural network.
      save(self, path: str) -> str
        A save of the neural network.
      create(self, input_size: int, num_classes: int, dropout: float) -> None
        Re creates the neural network.
  """
  def __call__(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
    """
    return self.forward(x)

  def __repr__(self):
    """
      A string representation of the neural network.

      Returns:
      --------
        : str
          A string representation 'Basic()'.
    """
    return 'Basic()'

  def __init__(self, *, data: dict|None = None, labels = [],
               input_size: int = 784, dropout: float = 0.0) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
        data: dict
          A dictionary containing the data, to load the network though the pth file.
        labels: list
          A list of labels.
        input_size: int
          The input size.
        dropout: float
          The dropout probability.
    """
    super().__init__()
    if data is None:
      self.labels = labels
      self.num_classes = len(labels)
      self.training_history = []

      self.input_size = input_size
      self.dropout = dropout
      last_size = self.__create_features__()
      self.__create_classifier__(last_size)
    else:
      if 'dropout' in data:
        self.dropout = data['dropout']
      self.labels = data['labels']
      self.num_classes = len(self.labels)
      self.input_size = data['input_size']
      self.features = data['features']
      if 'training_history' in data:
        self.training_history = data['training_history']
      if 'dropout' in data:
        self.dropout = data['dropout']
      if 'classifier' in data:
        self.classifier = data['classifier']

  def __create_classifier__(self, last_size: int) -> None:
    """
      Creates the classifier.

      Parameters:
      -----------
        last_size: int
          The size of the last layer.
    """
    self.classifier = torch.nn.Sequential(
      torch.nn.Linear(last_size, self.num_classes),
      torch.nn.LogSoftmax(dim=1)
    )

  def __create_features__(self) -> int:
    """
      Creates the features.

      Returns:
      --------
        int
          The size of the last features, to pipe it to the classifier.
    """
    last_size = 10
    if self.dropout > 0.0:
      self.features = torch.nn.Sequential(
        torch.nn.Linear(self.input_size, 128),
        torch.nn.ReLU(),
        torch.nn.Linear(128, 42),
        torch.nn.ReLU(),
        torch.nn.Dropout(self.dropout),
        torch.nn.Linear(42, 28),
        torch.nn.ReLU(),
        torch.nn.Dropout(self.dropout),
        torch.nn.Linear(28, 512),
        torch.nn.ReLU(),
        torch.nn.Dropout(self.dropout),
        torch.nn.Linear(512, 128),
        torch.nn.ReLU(),
        torch.nn.Dropout(self.dropout),
        torch.nn.Linear(128, 64),
        torch.nn.ReLU(),
        torch.nn.Linear(64, last_size),
        torch.nn.LogSoftmax(dim=1)
      )
    else:
      self.features = torch.nn.Sequential(
        torch.nn.Linear(self.input_size, 128),
        torch.nn.ReLU(),
        torch.nn.Linear(128, 42),
        torch.nn.ReLU(),
        torch.nn.Linear(42, 28),
        torch.nn.ReLU(),
        torch.nn.Linear(28, 512),
        torch.nn.ReLU(),
        torch.nn.Linear(512, 128),
        torch.nn.ReLU(),
        torch.nn.Linear(128, 64),
        torch.nn.ReLU(),
        torch.nn.Linear(64, last_size),
        torch.nn.LogSoftmax(dim=1)
      )
    return last_size

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor

      Help:
      -----
        model.forward = lambda x: model.classifier(model.features(x)).view(x.size(0), class_count)
    """
    x = self.features(x)
    x = self.classifier(x)
    return torch.flatten(x, 1)

  def parameters(self) -> Iterator[Parameter]:
    """
      Get the parameters of the neural network.

      Returns:
      --------
        Iterator[Parameter]
    """
    return self.features.parameters()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'dependencies': 'src/dependencies',
 'features': 'src/features',
 'features.comparer': 'src/features/comparer',
 'features.torchvision_downloader': 'src/features/torchvision_downloader',
 'features.trainer': 'src/features/trainer',
 'libs': 'src/libs',
 'models': 'src/modules/models',
 'modules': 'src/modules',
 'modules.models': 'src/modules/models',
 'modules.transforms': 'src/modules/transforms',
 'transforms': 'src/modules/transforms'}

packages = \
['commands',
 'dependencies',
 'features',
 'features.comparer',
 'features.torchvision_downloader',
 'features.trainer',
 'libs',
 'models',
 'modules',
 'modules.models',
 'modules.transforms',
 'transforms']

package_data = \
{'': ['*']}

modules = \
['main']
install_requires = \
['better-abc>=0.0.3,<0.0.4',
 'click>=8.1,<9.0',
 'loading-display>=0.2,<0.3',
 'matplotlib>=3.6,<4.0',
 'numpy>=1.24,<2.0',
 'pillow>=9.4,<10.0',
 'plotext>=5.2,<6.0',
 'termcolor>=2.2,<3.0',
 'torch>=1.13,<2.0',
 'torchvision>=0.14,<0.15',
 'wcmatch>=8.4,<9.0']

entry_points = \
{'console_scripts': ['alicia = main:call']}

setup_kwargs = {
    'name': 'alicia',
    'version': '0.0.13',
    'description': 'A CLI to download, train, test, predict and compare an image classifiers.',
    'long_description': '\n.. image:: https://img.shields.io/pypi/v/alicia.svg\n    :alt: PyPI-Server\n    :target: https://pypi.org/project/alicia/\n\n.. image:: https://pepy.tech/badge/alicia/month\n    :alt: Monthly Downloads\n    :target: https://pepy.tech/project/alicia\n\n\n================================================\n                   AlicIA\n================================================\n::\n\n  Usage: alicia [OPTIONS] COMMAND [ARGS]...\n\n    A CLI to download, train, test, predict and compare an image classifiers.\n\n    Supporting mostly all torch-vision neural networks and datasets.\n\n    This will also identify cute 🐱 or a fierce 🐶, also flowers or what type of\n    🏘️ you should be.\n\n  Options:\n    -v, --verbose\n    -g, --gpu\n    --version      Show the version and exit.\n    --help         Show this message and exit.\n\n  Commands:\n    compare   Compare the info, accuracy, and step speed two (or more by...\n    create    Creates a new model for a given architecture.\n    download  Download a MNIST dataset with PyTorch and split it into...\n    info      Display information about a model architecture.\n    predict   Predict images using a pre trained model, for a given folder...\n    test      Test a pre trained model.\n    train     Train a given architecture with a data directory containing a...\n\n\n.. image:: https://github.com/aemonge/alicia/raw/main/docs/DallE-Alicia-logo.png\n    :alt: DallE-Alicia-logo\n\nInstall and usage\n================================================\n::\n\n    pip install alicia\n    alicia --help\n\n\nIf you just want to see a quick showcase of the tool, download and run [showcase.sh](./showcase.sh)\n\nFeatures\n-----------------------------------------------\n\nTo see the full list of features, and option please refer to `alicia --help`\n\n* Download common torchvision datasets\n* Train, test and predict using different custom-made and torch-vision models.\n* Get information about each model.\n* Compare models training speed, accuracy, and meta information.\n* Tested with MNIST and FashionMNIST.\n* View results in the console, or with matplotlib\n\nReferences\n-----------------------------------------------\n\nUseful links found and used while developing this\n\n* (https://medium.com/analytics-vidhya/creating-a-custom-dataset-and-dataloader-in-pytorch-76f210a1df5d)\n* (https://stackoverflow.com/questions/51911749/what-is-the-difference-between-torch-tensor-and-torch-tensor)\n* (https://deepai.org/dataset/mnist)\n* (https://medium.com/fenwicks/tutorial-1-mnist-the-hello-world-of-deep-learning-abd252c47709)\n',
    'author': 'aemonge',
    'author_email': 'andres@aemonge.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/alicia/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import click
import contextlib
import csv
import difflib
import inspect
import io
import math
import multiprocessing
import numpy as np
import os
import pathlib
import random
import sys
import tempfile
import textwrap
import time
import torch
import torchvision

from PIL import Image
from random import randrange
from wcmatch import glob
from torchvision import transforms

from abc import abstractmethod
from better_abc import abstract_attribute, ABCMeta



from dependencies.core import csv

def labels_reader(file: str, _sorted: bool = True) -> (list|dict):
  """
    Reads labels from a CSV file.

    Parameters:
    -----------
      file: str
        Path to the CSV file.
      _sorted: bool
        Whether the labels should be sorted and returned as a list.

    Returns:
    --------
      list|dict
        Labels in the CSV file as a list or dictionary, dependent on the `sorted` parameter
  """
  if _sorted:
    labels = set()
  else:
    labels = {}

  with open(file, "r", encoding="utf-8") as f:
    reader = csv.reader(f)
    if _sorted:
      for _, label in reader:
        labels.add(label) # pyright: ignore [reportGeneralTypeIssues]
    else:
      for filename, label in reader:
        labels[filename] = label # pyright: ignore [reportGeneralTypeIssues]

  if _sorted:
    return sorted(list(labels)) # pyright: ignore [reportGeneralTypeIssues]
  return labels # pyright: ignore [reportGeneralTypeIssues]




from dependencies.core import click, torchvision
from features import TorchvisionDownloader

DATASETS=torchvision.datasets.__all__

@click.command()
@click.pass_context
@click.argument("dataset", default='MNIST', type=click.Choice(DATASETS))
@click.argument("data_dir", type=click.Path(dir_okay=True, readable=True), required=1)
@click.option("-s", "--split-percentage", default=(0.65, 0.25, 0.1), type=(float, float, float),
  help='The split percentage triplet "-s 0.65 0.25 0.1"'
)
@click.option("-f", "--force", type=click.BOOL, default=False, is_flag=True)
def download(_, dataset, data_dir, split_percentage, force):
  """
    Download a MNIST dataset with PyTorch and split it into `./train`, `./valid`, and `./test` directories.

    The download process will also generate the `./labels.csv` file containing the labels of all sets.
  """
  TorchvisionDownloader(data_dir, dataset, split_percentage).call(force)



from dependencies.core import time, random, torch
from dependencies.datatypes import Iterator, Parameter
from .abs_module import AbsModule

class Dummy(AbsModule):
  """
    A dummy module that return random sleep time, simple to test graphical output.

    Methods:
    --------
      forward(self, x: torch.Tensor) -> torch.Tensor
        A forward pass of the neural network.
  """
  def __call__(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
    """
    return self.forward(x)

  def __repr__(self):
    """
      A string representation of the neural network.

      Returns:
      --------
        : str
          A string representation 'Dummy()'.
    """
    return 'Dummy()'

  def __init__(self, _) -> None:
    """
      Constructor of the neural network.

      Parameters:
      -----------
    """
    super().__init__()
    self.num_classes = 5

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """
      A forward pass of the neural network.

      Parameters:
      -----------
        x: torch.Tensor
          A batch of input features.

      Returns:
      --------
        torch.Tensor
          Random tensor.
    """
    time.sleep(random.uniform(0.0, 2.5))
    return torch.randint(0, self.num_classes, (x.shape[0],))

  def load(self) -> None:
    """
      Not Implemented.

      Raises:
      ------
        NotImplementedError
    """
    raise NotImplementedError("A dummy is not expected to be used production")

  def parameters(self) -> Iterator[Parameter]:
    """
      Get the parameters of the neural network.

      Returns:
      --------
        Iterator[Parameter]
    """
    s = torch.nn.Sequential(torch.nn.Linear(10, self.num_classes))
    return s.parameters()

  def save(self) -> None:
    """
      Not Implemented.

      Raises:
      ------
        NotImplementedError
    """
    raise NotImplementedError("A dummy is not expected to be used production")

  def create(self) -> None:
    """
      Not Implemented.

      Raises:
      ------
        NotImplementedError
    """
    raise NotImplementedError("A dummy is not expected to be used production")
